from .object cimport PyObject

cdef extern from "Python.h":
    ############################################################################
    # 7.5.4 Method Objects
    ############################################################################

    # There are some useful functions that are useful for working with method objects.
    # PyTypeObject PyMethod_Type
    # This instance of PyTypeObject represents the Python method type. This is exposed to Python programs as types.MethodType.

    bint PyMethod_Check(object o)
    # Return true if o is a method object (has type
    # PyMethod_Type). The parameter must not be NULL.

    object PyMethod_New(object func, object self, object cls)
    # Return value: New reference.
    # Return a new method object, with func being any callable object;
    # this is the function that will be called when the method is
    # called. If this method should be bound to an instance, self
    # should be the instance and class should be the class of self,
    # otherwise self should be NULL and class should be the class
    # which provides the unbound method..

    PyObject* PyMethod_Class(object meth) except NULL
    # Return value: Borrowed reference.
    # Return the class object from which the method meth was created;
    # if this was created from an instance, it will be the class of
    # the instance.

    PyObject* PyMethod_GET_CLASS(object meth)
    # Return value: Borrowed reference.
    # Macro version of PyMethod_Class() which avoids error checking.

    PyObject* PyMethod_Function(object meth) except NULL
    # Return value: Borrowed reference.
    # Return the function object associated with the method meth.

    PyObject* PyMethod_GET_FUNCTION(object meth)
    # Return value: Borrowed reference.
    # Macro version of PyMethod_Function() which avoids error checking.

    PyObject* PyMethod_Self(object meth) except? NULL
    # Return value: Borrowed reference.
    # Return the instance associated with the method meth if it is bound, otherwise return NULL.

    PyObject* PyMethod_GET_SELF(object meth)
    # Return value: Borrowed reference.
    # Macro version of PyMethod_Self() which avoids error checking.



from Hmm.Hmm cimport Hmm
from Math.Vector cimport Vector
from Math.Matrix cimport Matrix


cdef class Hmm2(Hmm):

    cdef Matrix __pi

    def __init__(self,
                 states: set,
                 observations: list,
                 emittedSymbols: list):
        """
        A constructor of Hmm2 class which takes a Set of states, an array of observations (which also
        consists of an array of states) and an array of instances (which also consists of an array of emitted symbols).
        The constructor calls its super method to calculate the emission probabilities for those states.

        PARAMETERS
        ----------
        states : set
            A Set of states, consisting of all possible states for this problem.
        observations : list
            An array of instances, where each instance consists of an array of states.
        emittedSymbols : list
            An array of instances, where each instance consists of an array of symbols.
        """
        super().__init__(states,
                         observations,
                         emittedSymbols)

    cpdef calculatePi(self, list observations):
        """
        calculatePi calculates the prior probability matrix (initial probabilities for each state combinations)
        from a set of observations. For each observation, the function extracts the first and second states in
        that observation.  Normalizing the counts of the pair of states returns us the prior probabilities for each
        pair of states.

        PARAMETERS
        ----------
        observations : list
            A set of observations used to calculate the prior probabilities.
        """
        cdef int first, second
        cdef list observation
        self.__pi = Matrix(self.state_count, self.state_count)
        for observation in observations:
            first = self.state_indexes[observation[0]]
            second = self.state_indexes[observation[1]]
            self.__pi.increment(first, second)
        self.__pi.columnWiseNormalize()

    cpdef calculateTransitionProbabilities(self, list observations):
        """
        calculateTransitionProbabilities calculates the transition probabilities matrix from each state to another
        state. For each observation and for each transition in each observation, the function gets the states.
        Normalizing the counts of the three of states returns us the transition probabilities.

        PARAMETERS
        ----------
        observations : list
            A set of observations used to calculate the transition probabilities.
        """
        cdef list current
        cdef int j, from_index1, from_index2, to_index
        self.transition_probabilities = Matrix(self.state_count * self.state_count, self.state_count)
        for current in observations:
            for j in range(len(current) - 2):
                from_index1 = self.state_indexes[current[j]]
                from_index2 = self.state_indexes[current[j + 1]]
                to_index = self.state_indexes[current[j + 2]]
                self.transition_probabilities.increment(from_index1 * self.state_count + from_index2, to_index)
        self.transition_probabilities.columnWiseNormalize()

    cpdef Vector __logOfColumn(self, int column):
        """
        logOfColumn calculates the logarithm of each value in a specific column in the transition probability matrix.

        PARAMETERS
        ----------
        column : int
            Column index of the transition probability matrix.

        RETURNS
        -------
        Vector
            A vector consisting of the logarithm of each value in the column in the transition probability matrix.
        """
        cdef Vector result
        cdef int i
        result = Vector()
        for i in range(self.state_count):
            result.add(self.safeLog(self.transition_probabilities.getValue(i * self.state_count + column
                                                                          // self.state_count, column %
                                                                          self.state_count)))
        return result

    cpdef list viterbi(self, list s):
        """
        viterbi calculates the most probable state sequence for a set of observed symbols.

        PARAMETERS
        ----------
        s : list
            A set of observed symbols.

        RETURNS
        -------
        list
            The most probable state sequence as an {@link ArrayList}.
        """
        cdef list result
        cdef int sequence_length, i, j, t, max_index
        cdef Matrix gamma, phi
        cdef Vector qs, current, previous
        cdef double observation_likelihood
        result = []
        sequence_length = len(s)
        gamma = Matrix(sequence_length, self.state_count * self.state_count)
        phi = Matrix(sequence_length, self.state_count * self.state_count)
        qs = Vector(sequence_length, 0)
        emission1 = s[0]
        emission2 = s[1]
        for i in range(self.state_count):
            for j in range(self.state_count):
                observation_likelihood = self.states[i].getEmitProb(emission1) * self.states[j].getEmitProb(emission2)
                gamma.setValue(1, i * self.state_count + j, self.safeLog(self.__pi.getValue(i, j)) +
                               self.safeLog(observation_likelihood))
        for t in range(2, sequence_length):
            emission = s[t]
            for j in range(self.state_count * self.state_count):
                current = self.__logOfColumn(j)
                previous = gamma.getRowVector(t - 1).skipVector(self.state_count, j // self.state_count)
                current.addVector(previous)
                max_index = current.maxIndex()
                observation_likelihood = self.states[j % self.state_count].getEmitProb(emission)
                gamma.setValue(t, j, current.getValue(max_index) + self.safeLog(observation_likelihood))
                phi.setValue(t, j, max_index * self.state_count + j // self.state_count)
        qs.setValue(sequence_length - 1, gamma.getRowVector(sequence_length - 1).maxIndex())
        result.insert(0, self.states[int(qs.getValue(sequence_length - 1)) % self.state_count].getState())
        for i in range(sequence_length - 2, 0, -1):
            qs.setValue(i, phi.getValue(i + 1, int(qs.getValue(i + 1))))
            result.insert(0, self.states[int(qs.getValue(i)) % self.state_count].getState())
        result.insert(0, self.states[int(qs.getValue(1)) // self.state_count].getState())
        return result

    def __repr__(self):
        return f"{self.__pi} {self.transition_probabilities} {self.states}"



cdef class CounterHashMap(dict):

    cpdef put(self, object key)
    cpdef putNTimes(self, object key, int N)
    cpdef int count(self, object key)
    cpdef int sumOfCounts(self)
    cpdef object max(self, double threshold = *)
    cpdef add(self, CounterHashMap toBeAdded)
    cpdef list topN(self, int N)


from Math.Matrix cimport Matrix


cdef class Hmm(object):

    cdef Matrix transition_probabilities
    cdef dict state_indexes
    cdef list states
    cdef int state_count

    cpdef dict calculateEmissionProbabilities(self, object state, list observations, list emittedSymbols)
    cpdef double safeLog(self, double x)



from DataStructure.CounterHashMap cimport CounterHashMap
from Hmm.HmmState cimport HmmState
import math


cdef class Hmm(object):

    def calculatePi(self, observations: list):
        pass

    def calculateTransitionProbabilities(self, observations: list):
        pass

    def viterbi(self, s: list) -> list:
        pass

    def __init__(self,
                 states: set,
                 observations: list,
                 emittedSymbols: list):
        """
        A constructor of Hmm class which takes a Set of states, an array of observations (which also
        consists of an array of states) and an array of instances (which also consists of an array of emitted symbols).
        The constructor initializes the state array with the set of states and uses observations and emitted symbols
        to calculate the emission probabilities for those states.

        PARAMETERS
        ----------
        states : set
            A Set of states, consisting of all possible states for this problem.
        observations : list
            An array of instances, where each instance consists of an array of states.
        emittedSymbols : list
            An array of instances, where each instance consists of an array of symbols.
        """
        cdef int i
        cdef dict emission_probabilities
        i = 0
        self.state_count = len(states)
        self.states = []
        self.state_indexes = {}
        for state in states:
            self.state_indexes[state] = i
            i = i + 1
        self.calculatePi(observations)
        for state in states:
            emission_probabilities = self.calculateEmissionProbabilities(state, observations, emittedSymbols)
            self.states.append(HmmState(state, emission_probabilities))
        self.calculateTransitionProbabilities(observations)

    cpdef dict calculateEmissionProbabilities(self, object state, list observations, list emittedSymbols):
        """
        calculateEmissionProbabilities calculates the emission probabilities for a specific state. The method takes the
        state, an array of observations (which also consists of an array of states) and an array of instances (which also
        consists of an array of emitted symbols).

        PARAMETERS
        ----------
        states : set
            A Set of states, consisting of all possible states for this problem.
        observations : list
            An array of instances, where each instance consists of an array of states.
        emittedSymbols : list
            An array of instances, where each instance consists of an array of symbols.

        RETURNS
        -------
        dict
            A HashMap. Emission probabilities for a single state. Contains a probability for each symbol emitted.
        """
        cdef CounterHashMap counts
        cdef dict emission_probabilities
        cdef int i, j
        counts = CounterHashMap()
        emission_probabilities = {}
        for i in range(len(observations)):
            for j in range(len(observations[i])):
                current_state = observations[i][j]
                current_symbol = emittedSymbols[i][j]
                if current_state == state:
                    counts.put(current_symbol)
        total = counts.sumOfCounts()
        for symbol in counts:
            emission_probabilities[symbol] = counts[symbol] / total
        return emission_probabilities

    cpdef double safeLog(self, double x):
        """
        safeLog calculates the logarithm of a number. If the number is less than 0, the logarithm is not defined, therefore
        the function returns -Infinity.

        PARAMETERS
        ----------
        x : float
            Input number

        RETURNS
        -------
        float
            The logarithm of x. If x < 0 return -infinity.
        """
        if x <= 0:
            return -1000
        else:
            return math.log(x)

    def __repr__(self):
        return f"{self.transition_probabilities} {self.states}"



import math


cdef class DiscreteDistribution(dict):

    def __init__(self, **kwargs):
        """
        A constructor of DiscreteDistribution class which calls its super class.
        """
        super().__init__(**kwargs)
        self.__sum = 0.0

    cpdef addItem(self, str item):
        """
        The addItem method takes a String item as an input and if this map contains a mapping for the item it puts the
        item with given value + 1, else it puts item with value of 1.

        PARAMETERS
        ----------
        item : string
            String input.
        """
        if item in self:
            self[item] = self[item] + 1
        else:
            self[item] = 1
        self.__sum = self.__sum + 1

    cpdef removeItem(self, str item):
        """
        The removeItem method takes a String item as an input and if this map contains a mapping for the item it puts
        the item with given value - 1, and if its value is 0, it removes the item.

        PARAMETERS
        ----------
        item : string
            String input.
        """
        if item in self:
            self[item] = self[item] - 1
            if self[item] == 0:
                self.pop(item)

    cpdef addDistribution(self, DiscreteDistribution distribution):
        """
        The addDistribution method takes a DiscreteDistribution as an input and loops through the entries in this
        distribution and if this map contains a mapping for the entry it puts the entry with its value + entry,
        else it puts entry with its value. It also accumulates the values of entries and assigns to the sum variable.

        PARAMETERS
        ----------
        distribution : DiscreteDistribution
            DiscreteDistribution type input.
        """
        for entry in distribution:
            if entry in self:
                self[entry] = self[entry] + distribution[entry]
            else:
                self[entry] = distribution[entry]
            self.__sum += distribution[entry]

    cpdef removeDistribution(self, DiscreteDistribution distribution):
        """
        The removeDistribution method takes a DiscreteDistribution as an input and loops through the entries in this
        distribution and if this map contains a mapping for the entry it puts the entry with its key - value, else it
        removes the entry. It also decrements the value of entry from sum and assigns to the sum variable.

        PARAMETERS
        ----------
        distribution : DiscreteDistribution
            DiscreteDistribution type input.
        """
        for entry in distribution:
            if self[entry] - distribution[entry] != 0:
                self[entry] -= distribution[entry]
            else:
                self.pop(entry)
            self.__sum -= distribution[entry]

    cpdef double getSum(self):
        """
        The getter for sum variable.

        RETURNS
        -------
        double
            sum
        """
        return self.__sum

    cpdef int getIndex(self, str item):
        """
        The getIndex method takes an item as an input and returns the index of given item.

        PARAMETERS
        ----------
        item : string
            item to search for index.

        RETURNS
        -------
        int
            index of given item.
        """
        return list(self.keys()).index(item)

    cpdef bint containsItem(self, str item):
        """
        The containsItem method takes an item as an input and returns true if this map contains a mapping for the
        given item.

        PARAMETERS
        ----------
        item : string
            item to check.

        RETURNS
        -------
        boolean
            true if this map contains a mapping for the given item.
        """
        return item in self

    cpdef str getItem(self, int index):
        """
        The getItem method takes an index as an input and returns the item at given index.

        PARAMETERS
        ----------
        index : int
            index is used for searching the item.

        RETURNS
        -------
        string
            the item at given index.
        """
        return list(self.keys())[index]

    cpdef int getValue(self, int index):
        """
        The getValue method takes an index as an input and returns the value at given index.

        PARAMETERS
        ----------
        index : int
            index is used for searching the value.

        RETURNS
        -------
        int
            the value at given index.
        """
        return list(self.values())[index]

    cpdef int getCount(self, str item):
        """
        The getCount method takes an item as an input returns the value to which the specified item is mapped, or ""
        if this map contains no mapping for the key.

        PARAMETERS
        ----------
        item : string

        RETURNS
        -------
        int
            the value to which the specified item is mapped
        """
        return self[item]

    cpdef str getMaxItem(self):
        """
        The getMaxItem method loops through the entries and gets the entry with maximum value.

        RETURNS
        -------
        string
            the entry with maximum value.
        """
        cdef int maxValue
        cdef str maxItem
        maxValue = -1
        maxItem = ""
        for item in self:
            if self[item] > maxValue:
                maxValue = self[item]
                maxItem = item
        return maxItem

    cpdef str getMaxItemIncludeTheseOnly(self, list includeTheseOnly):
        """
        Another getMaxItem method which takes a list of Strings. It loops through the items in this list
        and gets the item with maximum value.

        PARAMETERS
        ----------
        includeTheseOnly : list
            list of Strings.

        RETURNS
        -------
        string
            the item with maximum value.
        """
        cdef int maxValue, frequency
        cdef str maxItem
        maxValue = -1
        maxItem = ""
        for item in includeTheseOnly:
            frequency = 0
            if item in self:
                frequency = self[item]
            if frequency > maxValue:
                maxValue = frequency
                maxItem = item
        return maxItem

    cpdef double getProbability(self, str item):
        """
        The getProbability method takes an item as an input returns the value to which the specified item is mapped over
        sum, or 0.0 if this map contains no mapping for the key.

        PARAMETERS
        ----------
        item : string
            is used to search for probability.

        RETURNS
        -------
        double
            the probability to which the specified item is mapped.
        """
        if item in self:
            return self[item] / self.__sum
        else:
            return 0.0

    cpdef dict getProbabilityDistribution(self):
        result = {}
        for item in self:
            result[item] = self.getProbability(item)
        return result

    cpdef double getProbabilityLaplaceSmoothing(self, str item):
        """
        The getProbabilityLaplaceSmoothing method takes an item as an input returns the smoothed value to which the
        specified item is mapped over sum, or 1.0 over sum if this map contains no mapping for the key.
        PARAMETERS
        ----------
        item : string
            is used to search for probability.

        RETURNS
        -------
        double
            the smoothed probability to which the specified item is mapped.
        """
        if item in self:
            return (self[item] + 1) / (self.__sum + len(self) + 1)
        else:
            return 1.0 / (self.__sum + len(self) + 1)

    cpdef double entropy(self):
        """
        The entropy method loops through the values and calculates the entropy of these values.

        RETURNS
        -------
        double
            entropy value.
        """
        cdef double total, probability
        cdef int count
        total = 0.0
        for count in self.values():
            probability = count / self.__sum
            total += -probability * math.log2(probability)
        return total



cdef extern from "Python.h":

    ###########################################################################
    # Iterator Objects
    ###########################################################################

    bint PySeqIter_Check(object op)
    # Return true if the type of op is PySeqIter_Type.

    object PySeqIter_New(object seq)
    # Return value: New reference.
    # Return an iterator that works with a general sequence object, seq. The
    # iteration ends when the sequence raises IndexError for the subscripting
    # operation.

    bint PyCallIter_Check(object op)
    # Return true if the type of op is PyCallIter_Type.

    object PyCallIter_New(object callable, object sentinel)
    # Return value: New reference.
    # Return a new iterator. The first parameter, callable, can be any Python
    # callable object that can be called with no parameters; each call to it
    # should return the next item in the iteration. When callable returns a
    # value equal to sentinel, the iteration will be terminated.



import math


cdef class Distribution(object):
    Z_MAX = 6.0
    Z_EPSILON = 0.000001
    CHI_EPSILON = 0.000001
    CHI_MAX = 99999.0
    LOG_SQRT_PI = 0.5723649429247000870717135
    I_SQRT_PI = 0.5641895835477562869480795
    BIGX = 200.0
    I_PI = 0.3183098861837906715377675
    F_EPSILON = 0.000001
    F_MAX = 9999.0

    @staticmethod
    def __ex(x: double) -> double:
        """
        The ex method takes a double x as an input, if x is less than -BIGX it returns 0, otherwise it returns Euler's
        number e raised to the power of x.

        PARAMETERS
        ----------
        x : double
            double input.

        RETURNS
        -------
        double
            0 if input is less than -BIGX, Euler's number e raised to the power of x otherwise.
        """
        if x < -Distribution.BIGX:
            return 0
        return math.exp(x)

    @staticmethod
    def beta(x: list) -> double:
        """
        The beta method takes a double list x as an input. It loops through x and accumulates
        the value of gammaLn(x), also it sums up the items of x and returns (accumulated result - gammaLn of this
        summation).

        PARAMETERS
        ----------
        x : list
            double list input.

        RETURNS
        -------
        double
            beta distribution at point x.
        """
        cdef double total, result
        cdef int i
        total = 0.0
        result = 0.0
        for i in range(len(x)):
            result += Distribution.gammaLn(x[i])
            total += x[i]
        result -= Distribution.gammaLn(total)
        return result

    @staticmethod
    def gammaLn(x: double) -> double:
        """
        The gammaLn method takes a double x as an input and returns the logarithmic result of the gamma distribution at
        point x.

        PARAMETERS
        ----------
        x : double
            double input.

        RETURNS
        -------
        double
            the logarithmic result of the gamma distribution at point x.
        """
        cdef double y, tmp, ser
        cdef int j
        cdef list cof
        cof = [76.18009172947146, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.1208650973866179e-2,
               -0.5395239384953e-5]
        y = x
        tmp = x + 5.5
        tmp -= (x + 0.5) * math.log(tmp)
        ser = 1.000000000190015
        for j in range(6):
            y = y + 1
            ser += cof[j] / y
        return -tmp + math.log(2.5066282746310005 * ser / x)

    @staticmethod
    def zNormal(z: double) -> double:
        """
        The zNormal method performs the Z-Normalization. It ensures, that all elements of the input vector are
        transformed into the output vector whose mean is approximately 0 while the standard deviation is in a range
        close to 1.

        PARAMETERS
        ----------
        z : double
            double input.

        RETURNS
        -------
        double
            normalized value of given input.
        """
        cdef double x, y, w
        if z == 0.0:
            x = 0.0
        else:
            y = 0.5 * abs(z)
            if y >= Distribution.Z_MAX * 0.5:
                x = 1.0
            else:
                if y < 1.0:
                    w = y * y
                    x = ((((((((0.000124818987 * w - 0.001075204047) * w + 0.005198775019) * w - 0.019198292004) * w
                             + 0.059054035642) * w - 0.151968751364) * w + 0.319152932694) * w - 0.531923007300) * w
                         + 0.797884560593) * y * 2.0
                else:
                    y -= 2.0
                    x = (((((((((((((-0.000045255659 * y + 0.000152529290) * y - 0.000019538132) * y
                                   - 0.000676904986) * y + 0.001390604284) * y - 0.000794620820) * y
                                - 0.002034254874) * y + 0.006549791214) * y - 0.010557625006) * y + 0.011630447319) * y
                            - 0.009279453341) * y + 0.005353579108) * y - 0.002141268741) * y + 0.000535310849) * y \
                        + 0.999936657524
        if z > 0.0:
            return (x + 1.0) * 0.5
        else:
            return (1.0 - x) * 0.5

    @staticmethod
    def zInverse(p: double) -> double:
        """
        The zInverse method returns the Z-Inverse of given probability value.

        PARAMETERS
        ----------
        p : double
            probability input.

        RETURNS
        -------
        double
            the Z-Inverse of given probability.
        """
        cdef double minz, maxz, zval, pval
        minz = -Distribution.Z_MAX
        maxz = Distribution.Z_MAX
        zval = 0.0
        if p <= 0.0 or p >= 1.0:
            return 0.0
        while maxz - minz > Distribution.Z_EPSILON:
            pval = Distribution.zNormal(zval)
            if pval > p:
                maxz = zval
            else:
                minz = zval
            zval = (maxz + minz) * 0.5
        return zval

    @staticmethod
    def chiSquare(x: double,
                  freedom: int) -> double:
        """
        The chiSquare method is used to determine whether there is a significant difference between the expected
        frequencies and the observed frequencies in one or more categories. It takes a double input x and an integer
        freedom for degrees of freedom as inputs. It returns the Chi Squared result.

        PARAMETERS
        ----------
        x : double
            double input.
        freedom : int
            integer input for degrees of freedom.

        RETURNS
        -------
        double
            the Chi Squared result.
        """
        cdef double y, a, s, z, e, c
        y = 0
        if x <= 0.0 or freedom < 1:
            return 1.0
        a = 0.5 * x
        even = (freedom % 2 == 0)
        if freedom > 1:
            y = Distribution.__ex(-a)
        if even:
            s = y
        else:
            s = 2.0 * Distribution.zNormal(-math.sqrt(x))
        if freedom > 2:
            x = 0.5 * (freedom - 1.0)
            if even:
                z = 1.0
            else:
                z = 0.5
            if a > Distribution.BIGX:
                if even:
                    e = 0.0
                else:
                    e = Distribution.LOG_SQRT_PI
                c = math.log(a)
                while z <= x:
                    e = math.log(z) + e
                    s += Distribution.__ex(c * z - a - e)
                    z += 1.0
                return s
            else:
                if even:
                    e = 1.0
                else:
                    e = Distribution.I_SQRT_PI / math.sqrt(a)
                c = 0.0
                while z <= x:
                    e = e * (a / z)
                    c = c + e
                    z += 1.0
                return c * y + s
        else:
            return s

    @staticmethod
    def chiSquareInverse(p: double,
                         freedom: int) -> double:
        """
        The chiSquareInverse method returns the Chi Square-Inverse of given probability value with given degree of
        freedom.

        PARAMETERS
        ----------
        p : double
            probability input.
        freedom : int
            integer input for degrees of freedom.

        RETURNS
        -------
        double
            the chiSquare-Inverse of given probability.
        """
        cdef double minchisq, maxchisq, chisqval
        minchisq = 0.0
        maxchisq = Distribution.CHI_MAX
        if p <= 0.0:
            return maxchisq
        else:
            if p >= 1.0:
                return 0.0
        chisqval = freedom / math.sqrt(p)
        while maxchisq - minchisq > Distribution.CHI_EPSILON:
            if Distribution.chiSquare(chisqval, freedom) < p:
                maxchisq = chisqval
            else:
                minchisq = chisqval
            chisqval = (maxchisq + minchisq) * 0.5
        return chisqval

    @staticmethod
    def fDistribution(F: double,
                      freedom1: int,
                      freedom2: int) -> double:
        """
        The fDistribution method is used to observe whether two samples have the same variance. It takes a double input
        F and two integer freedom1 and freedom2 for degrees of freedom as inputs. It returns the F-Distribution result.

        PARAMETERS
        ----------
        F : double
            double input.
        freedom1 : int
            integer input for degrees of freedom.
        freedom2 : int
            integer input for degrees of freedom.

        RETURNS
        -------
        double
            the F-Distribution result.
        """
        cdef double w, z, p, y, d
        cdef int a, b, j, i
        if F < Distribution.F_EPSILON or freedom1 < 1 or freedom2 < 1:
            return 1.0
        if freedom1 % 2 != 0:
            a = 1
        else:
            a = 2
        if freedom2 % 2 != 0:
            b = 1
        else:
            b = 2
        w = (F * freedom1) / freedom2
        z = 1.0 / (1.0 + w)
        if a == 1:
            if b == 1:
                p = math.sqrt(w)
                y = Distribution.I_PI
                d = y * z / p
                p = 2.0 * y * math.atan(p)
            else:
                p = math.sqrt(w * z)
                d = 0.5 * p * z / w
        else:
            if b == 1:
                p = math.sqrt(z)
                d = 0.5 * z * p
                p = 1.0 - p
            else:
                d = z * z
                p = w * z
        y = 2.0 * w / z
        for j in range(b + 2, freedom2 + 1, 2):
            d *= (1.0 + a / (j - 2.0)) * z
            if a == 1:
                p = p + d * y / (j - 1.0)
            else:
                p = (p + w) * z
        y = w * z
        z = 2.0 / z
        b = freedom2 - 2
        for i in range(a + 2, freedom1 + 1, 2):
            j = i + b
            d *= y * j / (i - 2.0)
            p -= z * d / j
        if p < 0.0:
            p = 0.0
        else:
            if p > 1.0:
                p = 1.0
        return 1.0 - p

    @staticmethod
    def fDistributionInverse(p: double,
                             freedom1: int,
                             freedom2: int) -> double:
        """
        The fDistributionInverse method returns the F-Distribution Inverse of given probability value.

        PARAMETERS
        ----------
        p : double
            double probability.
        freedom1 : int
            integer input for degrees of freedom.
        freedom2 : int
            integer input for degrees of freedom.

        RETURNS
        -------
        double
            the F-Distribution Inverse of given probability.
        """
        cdef double maxf, minf, fval
        maxf = Distribution.F_MAX
        minf = 0.0
        if p <= 0.0 or p >= 1.0:
            return 0.0
        if freedom1 == freedom2 and freedom1 > 2500:
            return 1 + 4.0 / freedom1
        fval = 1.0 / p
        while abs(maxf - minf) > Distribution.F_EPSILON:
            if Distribution.fDistribution(fval, freedom1, freedom2) < p:
                maxf = fval
            else:
                minf = fval
            fval = (maxf + minf) * 0.5
        return fval

    @staticmethod
    def tDistribution(T: double,
                      freedom: int) -> double:
        """
        The tDistribution method is used instead of the normal distribution when there is small samples. It takes a
        double input T and an integer freedom for degree of freedom as inputs. It returns the T-Distribution result by
        using F-Distribution method.

        PARAMETERS
        ----------
        T : double
            double input.
        freedom : int
            integer input for degrees of freedom.

        RETURNS
        -------
        double
            the T-Distribution result.
        """
        if T >= 0:
            return Distribution.fDistribution(T * T, 1, freedom) / 2
        else:
            return 1 - Distribution.fDistribution(T * T, 1, freedom) / 2

    @staticmethod
    def tDistributionInverse(p: double,
                             freedom: int) -> double:
        """
        The tDistributionInverse method returns the T-Distribution Inverse of given probability value.

        PARAMETERS
        ----------
        p : double
            double probability.
        freedom : int
            integer input for degrees of freedom.

        RETURNS
        -------
        double
            the T-Distribution Inverse of given probability.
        """
        if p < 0.5:
            return math.sqrt(Distribution.fDistributionInverse(p * 2, 1, freedom))
        else:
            return -math.sqrt(Distribution.fDistributionInverse((1 - p) * 2, 1, freedom))



import math

from Math.VectorSizeMismatch import VectorSizeMismatch


cdef class Vector(object):

    def __init__(self,
                 valuesOrSize=None,
                 initial=None):
        """
        A constructor of Vector class which takes a list values as an input. Then, initializes
        values list and size variable with given input and its size.

        PARAMETERS
        ----------
        valuesOrSize
            list input or size.
        initial
            initial value for each element
        """
        if valuesOrSize is None:
            self.__values = []
            self.__size = 0
        elif isinstance(valuesOrSize, list):
            self.__values = valuesOrSize.copy()
            self.__size = len(valuesOrSize)
        else:
            self.initAllSame(valuesOrSize, initial)

    cpdef initAllSame(self,
                      int size,
                      double x):
        """
        Another constructor of Vector class which takes integer size and double x as inputs. Then, initializes size
        variable with given size input and creates new values list and adds given input x to values list.

        PARAMETERS
        ----------
        size : int
            list size.
        x : double
            item to add values list.
        """
        cdef int i
        self.__size = size
        self.__values = []
        for i in range(size):
            self.__values.append(x)

    cpdef initAllZerosExceptOne(self,
                                int size,
                                int index,
                                double x):
        """
        Another constructor of Vector class which takes integer size, integer index and double x as inputs.
        Then, initializes size variable with given size input and creates new values list and adds 0.0 to
        values list. Then, sets the item of values list at given index as given input x.

        PARAMETERS
        ----------
        size : int
            list size.
        index : int
            to set a particular item.
        x : double
            item to add values list's given index.
        """
        cdef int i
        self.__size = size
        self.__values = []
        for i in range(size):
            self.__values.append(0.0)
        self.__values[index] = x

    cpdef Vector biased(self):
        """
        The biased method creates a list result, add adds each item of values list into the result list.
        Then, insert 1.0 to 0th position and return result list.

        RETURNS
        -------
        list
            result list.
        """
        cdef Vector result
        cdef double value
        result = Vector()
        for value in self.__values:
            result.add(value)
        result.insert(0, 1.0)
        return result

    cpdef add(self, double x):
        """
        The add method adds given input to the values {@link ArrayList} and increments the size variable by one.

        PARAMETERS
        ----------
        x : double
            input to add values list.
        """
        self.__values.append(x)
        self.__size = self.__size + 1

    cpdef insert(self,
                 int pos,
                 double x):
        """
        The insert method puts given input to the given index of values list and increments the size variable by one.

        PARAMETERS
        ----------
        pos : int
            index to insert input.
        x : double
            input to insert to given index of values list.
        """
        self.__values.insert(pos, x)
        self.__size = self.__size + 1

    cpdef remove(self, int pos):
        """
        The remove method deletes the item at given input position of values list and decrements the size variable by
        one.

        PARAMETERS
        ----------
        pos : int
            index to remove from values list.
        """
        self.__values.pop(pos)
        self.__size = self.__size - 1

    cpdef clear(self):
        """
        The clear method sets all the elements of values list to 0.
        """
        cdef int i
        for i in range(len(self.__values)):
            self.__values[i] = 0

    cpdef double sumOfElements(self):
        """
        The sumOfElements method sums up all elements in the vector.

        RETURNS
        -------
        double
            Sum of all elements in the vector.
        """
        cdef double total
        cdef int i
        total = 0
        for i in range(self.__size):
            total += self.__values[i]
        return total

    cpdef int maxIndex(self):
        """
        The maxIndex method gets the first item of values list as maximum item, then it loops through the indices
        and if a greater value than the current maximum item comes, it updates the maximum item and returns the final
        maximum item's index.

        RETURNS
        -------
        int
            final maximum item's index.
        """
        cdef int index, i
        cdef double maxValue
        index = 0
        maxValue = self.__values[0]
        for i in range(1, self.__size):
            if self.__values[i] > maxValue:
                maxValue = self.__values[i]
                index = i
        return index

    cpdef sigmoid(self):
        """
        The sigmoid method loops through the values list and sets each ith item with sigmoid function, i.e
        1 / (1 + Math.exp(-values.get(i))), i ranges from 0 to size.
        """
        cdef int i
        for i in range(self.__size):
            self.__values[i] = 1 / (1 + math.exp(-self.__values[i]))

    cpdef tanh(self):
        """
        The tanh method loops through the values list and sets each ith item with tanh function.
        """
        cdef int i
        for i in range(self.__size):
            self.__values[i] = math.tanh(self.__values[i])

    cpdef relu(self):
        """
        The relu method loops through the values list and sets each ith item with relu function.
        """
        cdef int i
        for i in range(self.__size):
            if self.__values[i] < 0:
                self.__values[i] = 0.0

    cpdef reluDerivative(self):
        """
        The reluDerivative method loops through the values list and sets each ith item with the derivative of the
        relu function.
        """
        cdef int i
        for i in range(self.__size):
            if self.__values[i] > 0:
                self.__values[i] = 1.0
            else:
                self.__values[i] = 0.0

    cpdef Vector skipVector(self,
                            int mod,
                            int value):
        """
        The skipVector method takes a mod and a value as inputs. It creates a new result Vector, and assigns given input
        value to i. While i is less than the size, it adds the ith item of values {@link ArrayList} to the result and
        increments i by given mod input.

        PARAMETERS
        ----------
        mod : int
            integer input.
        value : int
            integer input.

        RETURNS
        -------
        Vector
            result Vector.
        """
        cdef Vector result
        cdef int i
        result = Vector()
        i = value
        while i < self.__size:
            result.add(self.__values[i])
            i += mod
        return result

    cpdef addVector(self, Vector v):
        """
        The add method takes a Vector v as an input. It sums up the corresponding elements of both given vector's
        values list and values list and puts result back to the values list.

        PARAMETERS
        ----------
        v : Vector
            Vector to add.
        """
        cdef int i
        if self.__size != v.__size:
            raise VectorSizeMismatch
        for i in range(self.__size):
            self.__values[i] = self.__values[i] + v.__values[i]

    cpdef subtract(self, Vector v):
        """
        The subtract method takes a Vector v as an input. It subtracts the corresponding elements of given vector's
        values list from values list and puts result back to the values list.

        PARAMETERS
        ----------
        v : Vector
            Vector to subtract from values list.
        """
        cdef int i
        if self.__size != v.__size:
            raise VectorSizeMismatch
        for i in range(self.__size):
            self.__values[i] = self.__values[i] - v.__values[i]

    cpdef Vector difference(self, Vector v):
        """
        The difference method takes a Vector v as an input. It creates a new Vector result, then
        subtracts the corresponding elements of given vector's values list from values list and puts
        result back to the result.

        PARAMETERS
        ----------
        v : Vector
            Vector to find difference from values list.

        RETURNS
        -------
        Vector
            new Vector with result list.
        """
        cdef int i
        if self.__size != v.__size:
            raise VectorSizeMismatch
        result = Vector()
        for i in range(self.__size):
            result.add(self.__values[i] - v.__values[i])
        return result

    cpdef double dotProduct(self, Vector v):
        """
        The dotProduct method takes a Vector v as an input. It creates a new double variable result, then
        multiplies the corresponding elements of given vector's values list with values list and assigns
        the multiplication to the result.

        PARAMETERS
        ----------
        v : Vector
            Vector to find dot product.

        RETURNS
        -------
        double
            result.
        """
        cdef double result
        cdef int i
        if self.__size != v.__size:
            raise VectorSizeMismatch
        result = 0
        for i in range(self.__size):
            result += self.__values[i] * v.__values[i]
        return result

    cpdef double dotProductWithSelf(self):
        """
        The dotProduct method creates a new double variable result, then squares the elements of values list and assigns
        the accumulation to the result.

        RETURNS
        -------
        double
            result.
        """
        cdef double result
        cdef int i
        result = 0
        for i in range(self.__size):
            result += self.__values[i] * self.__values[i]
        return result

    cpdef Vector elementProduct(self, Vector v):
        """
        The elementProduct method takes a Vector v as an input. It creates a new Vector result, then
        multiplies the corresponding elements of given vector's values list with values list and adds
        the multiplication to the result list.

        PARAMETERS
        ----------
        v : Vector
            Vector to find dot product.

        RETURNS
        -------
        Vector
            with result list.
        """
        cdef Vector result
        cdef int i
        if self.__size != v.__size:
            raise VectorSizeMismatch
        result = Vector()
        for i in range(self.__size):
            result.add(self.__values[i] * v.__values[i])
        return result

    cpdef divide(self, double value):
        """
        The divide method takes a double value as an input and divides each item of values list with given value.

        PARAMETERS
        ----------
        value : double
            is used to divide items of values list.
        """
        cdef int i
        for i in range(self.__size):
            self.__values[i] = self.__values[i] / value

    cpdef multiply(self, double value):
        """
        The multiply method takes a double value as an input and multiplies each item of values list with given value.

        PARAMETERS
        ----------
        value : double
            is used to multiply items of values list.
        """
        cdef int i
        for i in range(self.__size):
            self.__values[i] = self.__values[i] * value

    cpdef Vector product(self, double value):
        """
        The product method takes a double value as an input and creates a new result {@link Vector}, then multiplies
        each item of values list with given value and adds to the result {@link Vector}.

        PARAMETERS
        ----------
        value : double
            is used to multiply items of values list.

        RETURNS
        -------
        Vector
            Vector result.
        """
        cdef Vector result
        cdef int i
        result = Vector()
        for i in range(self.__size):
            result.add(self.__values[i] * value)
        return result

    cpdef l1Normalize(self):
        """
        The l1Normalize method is used to apply Least Absolute Errors, it accumulates items of values list and sets
        each item by dividing it by the summation value.
        """
        cdef double total
        cdef int i
        total = 0
        for i in range(self.__size):
            total += self.__values[i]
        for i in range(self.__size):
            self.__values[i] = self.__values[i] / total

    cpdef double l2Norm(self):
        """
        The l2Norm method is used to apply Least Squares, it accumulates second power of each items of values list
        and returns the square root of this summation.

        RETURNS
        -------
        double
            square root of this summation.
        """
        cdef double total
        cdef int i
        total = 0
        for i in range(self.__size):
            total += self.__values[i] ** 2
        return math.sqrt(total)

    cpdef double cosineSimilarity(self, Vector v):
        """
        The cosineSimilarity method takes a Vector v as an input and returns the result of dotProduct(v)
        / l2Norm() / v.l2Norm().

        PARAMETERS
        ----------
        v : Vector
            input.

        RETURNS
        -------
        double
            dotProduct(v) / l2Norm() / v.l2Norm()
        """
        if self.__size != v.__size:
            raise VectorSizeMismatch
        return self.dotProduct(v) / self.l2Norm() / v.l2Norm()

    cpdef int size(self):
        """
        The size method returns the size of the values list

        RETURNS
        -------
        int
            size of the values list
        """
        return len(self.__values)

    cpdef double getValue(self, int index):
        """
        Getter for the item at given index of values list

        PARAMETERS
        ----------
        index : int
            index used to get an item.

        RETURNS
        -------
        item
            the item at given index.
        """
        return self.__values[index]

    cpdef setValue(self,
                   int index,
                   double value):
        """
        Setter for the setting the value at given index of values list.

        PARAMETERS
        ----------
        index : int
            index to set.
        value : item
            is used to set the given index
        """
        self.__values[index] = value

    cpdef addValue(self,
                   int index,
                   double value):
        """
        The addValue method adds the given value to the item at given index of values list.

        PARAMETERS
        ----------
        index : int
            index to add the given value.
        value : item
            value to add to given index.
        """
        self.__values[index] += value

    def __repr__(self):
        return f"{self.__values}"




cdef extern from "Python.h":
    ctypedef long long PY_LONG_LONG
    ctypedef unsigned long long uPY_LONG_LONG "unsigned PY_LONG_LONG"

    ############################################################################
    # 7.2.3 Long Integer Objects
    ############################################################################

    # PyLongObject
    #
    # This subtype of PyObject represents a Python long integer object.

    # PyTypeObject PyLong_Type
    #
    # This instance of PyTypeObject represents the Python long integer
    # type. This is the same object as long and types.LongType.

    bint PyLong_Check(object p)
    # Return true if its argument is a PyLongObject or a subtype of PyLongObject.

    bint PyLong_CheckExact(object p)
    # Return true if its argument is a PyLongObject, but not a subtype of PyLongObject.

    object PyLong_FromLong(long v)
    # Return value: New reference.
    # Return a new PyLongObject object from v, or NULL on failure.

    object PyLong_FromUnsignedLong(unsigned long v)
    # Return value: New reference.
    # Return a new PyLongObject object from a C unsigned long, or NULL on failure.

    object PyLong_FromSsize_t(Py_ssize_t v)
    # Return value: New reference.
    # Return a new PyLongObject object from a C Py_ssize_t, or NULL on failure.)

    object PyLong_FromSize_t(size_t v)
    # Return value: New reference.
    # Return a new PyLongObject object from a C size_t, or NULL on failure.

    object PyLong_FromLongLong(PY_LONG_LONG v)
    # Return value: New reference.
    # Return a new PyLongObject object from a C long long, or NULL on failure.

    object PyLong_FromUnsignedLongLong(uPY_LONG_LONG v)
    # Return value: New reference.
    # Return a new PyLongObject object from a C unsigned long long, or NULL on failure.

    object PyLong_FromDouble(double v)
    # Return value: New reference.
    # Return a new PyLongObject object from the integer part of v, or NULL on failure.

    object PyLong_FromString(char *str, char **pend, int base)
    # Return value: New reference.
    # Return a new PyLongObject based on the string value in str,
    # which is interpreted according to the radix in base. If pend is
    # non-NULL, *pend will point to the first character in str which
    # follows the representation of the number. If base is 0, the
    # radix will be determined based on the leading characters of str:
    # if str starts with '0x' or '0X', radix 16 will be used; if str
    # starts with '0', radix 8 will be used; otherwise radix 10 will
    # be used. If base is not 0, it must be between 2 and 36,
    # inclusive. Leading spaces are ignored. If there are no digits,
    # ValueError will be raised.

    object PyLong_FromUnicode(Py_UNICODE *u, Py_ssize_t length, int base)
    # Return value: New reference.
    # Convert a sequence of Unicode digits to a Python long integer
    # value. The first parameter, u, points to the first character of
    # the Unicode string, length gives the number of characters, and
    # base is the radix for the conversion. The radix must be in the
    # range [2, 36]; if it is out of range, ValueError will be
    # raised.

    # object PyLong_FromUnicodeObject(object u, int base)
    # Convert a sequence of Unicode digits in the string u to a Python integer
    # value. The Unicode string is first encoded to a byte string using
    # PyUnicode_EncodeDecimal() and then converted using PyLong_FromString().
    # New in version 3.3.

    object PyLong_FromVoidPtr(void *p)
    # Return value: New reference.
    # Create a Python integer or long integer from the pointer p. The
    # pointer value can be retrieved from the resulting value using
    # PyLong_AsVoidPtr().  If the integer is larger than LONG_MAX, a
    # positive long integer is returned.

    long PyLong_AsLong(object pylong) except? -1
    # Return a C long representation of the contents of pylong. If
    # pylong is greater than LONG_MAX, an OverflowError is raised.

    # long PyLong_AsLongAndOverflow(object pylong, int *overflow) except? -1
    # Return a C long representation of the contents of pylong. If pylong is
    # greater than LONG_MAX or less than LONG_MIN, set *overflow to 1 or -1,
    # respectively, and return -1; otherwise, set *overflow to 0. If any other
    # exception occurs (for example a TypeError or MemoryError), then -1 will
    # be returned and *overflow will be 0.
    # New in version 2.7.

    # PY_LONG_LONG PyLong_AsLongLongAndOverflow(object pylong, int *overflow) except? -1
    # Return a C long long representation of the contents of pylong. If pylong
    # is greater than PY_LLONG_MAX or less than PY_LLONG_MIN, set *overflow to
    # 1 or -1, respectively, and return -1; otherwise, set *overflow to 0. If
    # any other exception occurs (for example a TypeError or MemoryError), then
    # -1 will be returned and *overflow will be 0.
    # New in version 2.7.

    Py_ssize_t PyLong_AsSsize_t(object pylong) except? -1
    # Return a C Py_ssize_t representation of the contents of pylong. If pylong
    # is greater than PY_SSIZE_T_MAX, an OverflowError is raised and -1 will be
    # returned.

    unsigned long PyLong_AsUnsignedLong(object pylong) except? -1
    # Return a C unsigned long representation of the contents of
    # pylong. If pylong is greater than ULONG_MAX, an OverflowError is
    # raised.

    PY_LONG_LONG PyLong_AsLongLong(object pylong) except? -1
    # Return a C long long from a Python long integer. If pylong
    # cannot be represented as a long long, an OverflowError will be
    # raised.

    uPY_LONG_LONG PyLong_AsUnsignedLongLong(object pylong) except? -1
    #unsigned PY_LONG_LONG PyLong_AsUnsignedLongLong(object pylong)
    # Return a C unsigned long long from a Python long integer. If
    # pylong cannot be represented as an unsigned long long, an
    # OverflowError will be raised if the value is positive, or a
    # TypeError will be raised if the value is negative.

    unsigned long PyLong_AsUnsignedLongMask(object io) except? -1
    # Return a C unsigned long from a Python long integer, without
    # checking for overflow.

    uPY_LONG_LONG PyLong_AsUnsignedLongLongMask(object io) except? -1
    #unsigned PY_LONG_LONG PyLong_AsUnsignedLongLongMask(object io)
    # Return a C unsigned long long from a Python long integer,
    # without checking for overflow.

    double PyLong_AsDouble(object pylong) except? -1.0
    # Return a C double representation of the contents of pylong. If
    # pylong cannot be approximately represented as a double, an
    # OverflowError exception is raised and -1.0 will be returned.

    void* PyLong_AsVoidPtr(object pylong) except? NULL
    # Convert a Python integer or long integer pylong to a C void
    # pointer. If pylong cannot be converted, an OverflowError will be
    # raised. This is only assured to produce a usable void pointer
    # for values created with PyLong_FromVoidPtr(). For values outside
    # 0..LONG_MAX, both signed and unsigned integers are accepted.



cdef class Vector(object):

    cdef int __size
    cdef list __values

    cpdef initAllSame(self, int size, double x)
    cpdef initAllZerosExceptOne(self, int size, int index, double x)
    cpdef Vector biased(self)
    cpdef add(self, double x)
    cpdef insert(self, int pos, double x)
    cpdef remove(self, int pos)
    cpdef clear(self)
    cpdef double sumOfElements(self)
    cpdef int maxIndex(self)
    cpdef sigmoid(self)
    cpdef tanh(self)
    cpdef relu(self)
    cpdef reluDerivative(self)
    cpdef Vector skipVector(self, int mod, int value)
    cpdef addVector(self, Vector v)
    cpdef subtract(self, Vector v)
    cpdef Vector difference(self, Vector v)
    cpdef double dotProduct(self, Vector v)
    cpdef double dotProductWithSelf(self)
    cpdef Vector elementProduct(self, Vector v)
    cpdef divide(self, double value)
    cpdef multiply(self, double value)
    cpdef Vector product(self, double value)
    cpdef l1Normalize(self)
    cpdef double l2Norm(self)
    cpdef double cosineSimilarity(self, Vector v)
    cpdef int size(self)
    cpdef double getValue(self, int index)
    cpdef setValue(self, int index, double value)
    cpdef addValue(self, int index, double value)


from Math.Vector cimport Vector


cdef class Matrix(object):

    cdef int __row, __col
    cdef list __values

    cpdef initZeros(self)
    cpdef Matrix clone(self)
    cpdef double getValue(self, int rowNo, int colNo)
    cpdef setValue(self, int rowNo, int colNo, double value)
    cpdef addValue(self, int rowNo, int colNo, double value)
    cpdef increment(self, int rowNo, int colNo)
    cpdef int getRow(self)
    cpdef Vector getRowVector(self, int row)
    cpdef int getColumn(self)
    cpdef list getColumnVector(self, int column)
    cpdef columnWiseNormalize(self)
    cpdef multiplyWithConstant(self, double constant)
    cpdef divideByConstant(self, double constant)
    cpdef add(self, Matrix m)
    cpdef addRowVector(self, int rowNo, Vector v)
    cpdef subtract(self, Matrix m)
    cpdef Vector multiplyWithVectorFromLeft(self, Vector v)
    cpdef Vector multiplyWithVectorFromRight(self, Vector v)
    cpdef double columnSum(self, int columnNo)
    cpdef Vector sumOfRows(self)
    cpdef double rowSum(self, int rowNo)
    cpdef Matrix multiply(self, Matrix m)
    cpdef Matrix elementProduct(self, Matrix m)
    cpdef double sumOfElements(self)
    cpdef double trace(self)
    cpdef Matrix transpose(self)
    cpdef Matrix partial(self, int rowStart, int rowEnd, int colStart, int colEnd)
    cpdef bint isSymmetric(self)
    cpdef double determinant(self)
    cpdef inverse(self)
    cpdef Matrix choleskyDecomposition(self)
    cpdef __rotate(self, double s, double tau, int i, int j, int k, int l)


cdef class Stack:

    def __init__(self):
        self.__stack = []

    cpdef push(self, object item):
        self.__stack.append(item)

    cpdef object pop(self):
        if len(self.__stack) > 0:
            return self.__stack.pop()
        else:
            return None

    cpdef bint isEmpty(self):
        return len(self.__stack) == 0



import copy
import random
import math
from Math.MatrixNotPositiveDefinite import MatrixNotPositiveDefinite
from Math.MatrixNotSquare import MatrixNotSquare
from Math.Eigenvector import Eigenvector
from Math.MatrixDimensionMismatch import MatrixDimensionMismatch
from Math.Vector cimport Vector
from Math.MatrixRowMismatch import MatrixRowMismatch
from Math.DeterminantZero import DeterminantZero
from Math.MatrixRowColumnMismatch import MatrixRowColumnMismatch
from Math.MatrixColumnMismatch import MatrixColumnMismatch
from Math.MatrixNotSymmetric import MatrixNotSymmetric


cdef class Matrix(object):

    def __init__(self,
                 row,
                 col=None,
                 minValue=None,
                 maxValue=None,
                 seed=None):
        """
        Constructor of Matrix class which takes row and column numbers (Vectors) as inputs.

        PARAMETERS
        ----------
        row : int (or Vector)
            is used to create matrix.
        col : int (or Vector)
            is used to create matrix.
        minValue : double
            minimum Value for the initialization
        maxValue : double
            maximum Value for the initialization
        seed : int
            seed for the random
        """
        cdef int i, j
        if isinstance(row, int):
            self.__row = row
            if col is not None:
                self.__col = col
                if minValue is None:
                    self.initZeros()
                elif maxValue is None:
                    self.initZeros()
                    for i in range(self.__row):
                        self.__values[i][i] = minValue
                else:
                    random.seed(seed)
                    self.__values = [[random.uniform(minValue, maxValue) for _ in range(self.__col)] for _ in
                                     range(self.__row)]
            else:
                self.__col = row
                self.initZeros()
                for i in range(self.__row):
                    self.__values[i][i] = 1.0
        elif isinstance(row, Vector) and isinstance(col, Vector):
            self.__row = row.size()
            self.__col = col.size()
            self.initZeros()
            for i in range(row.size()):
                for j in range(col.size()):
                    self.__values[i][j] = row.getValue(i) * col.getValue(j)

    cpdef initZeros(self):
        cdef int i, j
        self.__values = [[0 for i in range(self.__col)] for j in range(self.__row)]

    cpdef Matrix clone(self):
        return copy.deepcopy(self)

    cpdef double getValue(self,
                          int rowNo,
                          int colNo):
        """
        The getter for the index at given rowNo and colNo of values list.

        PARAMETERS
        ----------
        rowNo : int
            integer input for row number.
        colNo : int
            integer input for column number.

        RETURNS
        -------
        double
            item at given index of values list.
        """
        return self.__values[rowNo][colNo]

    cpdef setValue(self,
                   int rowNo,
                   int colNo,
                   double value):
        """
        The setter for the value at given index of values list.

        PARAMETERS
        ----------
        rowNo : int
            integer input for row number.
        colNo : int
            integer input for column number.
        value : double
            is used to set at given index.
        """
        self.__values[rowNo][colNo] = value

    cpdef addValue(self,
                   int rowNo,
                   int colNo,
                   double value):
        """
        The addValue method adds the given value to the item at given index of values list.

        PARAMETERS
        ----------
        rowNo : int
            integer input for row number.
        colNo : int
            integer input for column number.
        value : double
            is used to add to given item at given index.
        """
        self.__values[rowNo][colNo] += value

    cpdef increment(self,
                    int rowNo,
                    int colNo):
        """
        The increment method adds 1 to the item at given index of values list.

        PARAMETERS
        ----------
        rowNo : int
            integer input for row number.
        colNo : int
            integer input for column number.
        """
        self.__values[rowNo][colNo] += 1

    cpdef int getRow(self):
        """
        The getter for the row variable.

        RETURNS
        -------
        int
            row number.
        """
        return self.__row

    cpdef Vector getRowVector(self, int row):
        """
        The getRowVector method returns the vector of values list at given row input.

        PARAMETERS
        ----------
        row : int
            row integer input for row number.

        RETURNS
        -------
        Vector
            Vector of values list at given row input.
        """
        cdef list rowList
        cdef Vector rowVector
        rowList = self.__values[row]
        rowVector = Vector(rowList)
        return rowVector

    cpdef int getColumn(self):
        """
        The getter for the col variable.

        RETURNS
        -------
        int
            column number.
        """
        return self.__col

    cpdef list getColumnVector(self, int column):
        """
         * The getColumnVector method creates a Vector and adds items at given column number of values list
         * to the Vector.

        PARAMETERS
        ----------
        column : int
            column integer input for column number.

        RETURNS
        -------
        Vector
            Vector of given column number.
        """
        cdef list columnVector
        cdef int i
        columnVector = []
        for i in range(self.__row):
            columnVector.append(self.__values[i][column])
        return columnVector

    cpdef columnWiseNormalize(self):
        """
        The columnWiseNormalize method, first accumulates items column by column then divides items
        by the summation.
        """
        cdef int i
        for i in range(self.__row):
            total = sum(self.__values[i])
            self.__values[i][:] = [x / total for x in self.__values[i]]

    cpdef multiplyWithConstant(self, double constant):
        """
        The multiplyWithConstant method takes a constant as an input and multiplies each item of values list
        with given constant.

        PARAMETERS
        ----------
        constant : double
            constant value to multiply items of values list.
        """
        cdef int i
        for i in range(self.__row):
            self.__values[i][:] = [x * constant for x in self.__values[i]]

    cpdef divideByConstant(self, double constant):
        """
        The divideByConstant method takes a constant as an input and divides each item of values list
        with given constant.

        PARAMETERS
        ----------
        constant : double
            constant value to divide items of values list.
        """
        cdef int i
        for i in range(self.__row):
            self.__values[i][:] = [x / constant for x in self.__values[i]]

    cpdef add(self, Matrix m):
        """
        The add method takes a Matrix as an input and accumulates values list with the
        corresponding items of given Matrix. If the sizes of both Matrix and values list do not match,
        it throws MatrixDimensionMismatch exception.

        PARAMETERS
        ----------
        m : Matrix
            Matrix type input.
        """
        cdef int i, j
        if self.__row != m.__row or self.__col != m.__col:
            raise MatrixDimensionMismatch
        for i in range(self.__row):
            for j in range(self.__col):
                self.__values[i][j] += m.__values[i][j]

    cpdef addRowVector(self,
                       int rowNo,
                       Vector v):
        """
        The add method which takes a row number and a Vector as inputs. It sums up the corresponding values at the given
        row of values list and given Vector. If the sizes of both Matrix and values list do not match, it throws
        MatrixColumnMismatch exception.

        PARAMETERS
        ----------
        rowNo : int
            integer input for row number.
        v : Vector
            Vector type input.
        """
        cdef int i
        if self.__col != v.size():
            raise MatrixColumnMismatch
        for i in range(self.__col):
            self.__values[rowNo][i] += v.getValue(i)

    cpdef subtract(self, Matrix m):
        """
        The subtract method takes a Matrix as an input and subtracts from values list the
        corresponding items of given Matrix. If the sizes of both Matrix and values list do not match,
        it throws {@link MatrixDimensionMismatch} exception.

        PARAMETERS
        ----------
        m : Matrix
            Matrix type input.
        """
        cpdef int i, j
        if self.__row != m.__row or self.__col != m.__col:
            raise MatrixDimensionMismatch
        for i in range(self.__row):
            for j in range(self.__col):
                self.__values[i][j] -= m.__values[i][j]

    cpdef Vector multiplyWithVectorFromLeft(self, Vector v):
        """
        The multiplyWithVectorFromLeft method takes a Vector as an input and creates a result list.
        Then, multiplies values of input Vector starting from the left side with the values list,
        accumulates the multiplication, and assigns to the result list. If the sizes of both Vector
        and row number do not match, it throws MatrixRowMismatch exception.

        PARAMETERS
        ----------
        v : Vector
            Vector type input.

        RETURNS
        -------
        Vector
            Vector that holds the result.
        """
        cdef int i, j
        cdef double total
        cdef Vector result
        if self.__row != v.size():
            raise MatrixRowMismatch
        result = Vector()
        for i in range(self.__col):
            total = 0.0
            for j in range(self.__row):
                total += v.getValue(j) * self.__values[j][i]
            result.add(total)
        return result

    cpdef Vector multiplyWithVectorFromRight(self, Vector v):
        """
        The multiplyWithVectorFromRight method takes a Vector as an input and creates a result list.
        Then, multiplies values of input Vector starting from the right side with the values list,
        accumulates the multiplication, and assigns to the result list. If the sizes of both Vector
        and row number do not match, it throws MatrixColumnMismatch exception.

        PARAMETERS
        ----------
        v : Vector
            Vector type input.

        RETURNS
        -------
        Vector
            Vector that holds the result.
        """
        cdef int i, j
        cdef double total
        cdef Vector result
        if self.__col != v.size():
            raise MatrixColumnMismatch
        result = Vector()
        for i in range(self.__row):
            total = 0.0
            for j in range(self.__col):
                total += v.getValue(j) * self.__values[i][j]
            result.add(total)
        return result

    cpdef double columnSum(self, int columnNo):
        """
        The columnSum method takes a column number as an input and accumulates items at given column number of values
        list.

        PARAMETERS
        ----------
        columnNo : int
            Column number input.

        RETURNS
        -------
        double
            summation of given column of values list.
        """
        cdef double total
        cdef int i
        total = 0
        for i in range(self.__row):
            total += self.__values[i][columnNo]
        return total

    cpdef Vector sumOfRows(self):
        """
        The sumOfRows method creates a mew result Vector and adds the result of columnDum method's corresponding
        index to the newly created result Vector.

        RETURNS
        -------
        Vector
            Vector that holds column sum.
        """
        cdef Vector result
        cdef int i
        result = Vector()
        for i in range(self.__col):
            result.add(self.columnSum(i))
        return result

    cpdef double rowSum(self, int rowNo):
        """
        The rowSum method takes a row number as an input and accumulates items at given row number of values list.

         * @param rowNo Row number input.
         * @return summation of given row of values {@link java.lang.reflect.Array}.
        """
        return sum(self.__values[rowNo])

    cpdef Matrix multiply(self, Matrix m):
        """
        The multiply method takes a Matrix as an input. First it creates a result Matrix and puts the
        accumulatated multiplication of values list and given Matrix into result
        Matrix. If the size of Matrix's row size and values list's column size do not match,
        it throws MatrixRowColumnMismatch exception.

        PARAMETERS
        ----------
        m : Matrix
            Matrix type input.

        RETURNS
        -------
        Matrix
            result Matrix.
        """
        cdef Matrix result
        cdef int i, j, k
        cdef double total
        if self.__col != m.__row:
            raise MatrixRowColumnMismatch
        result = Matrix(self.__row, m.__col)
        for i in range(self.__row):
            for j in range(m.__col):
                total = 0.0
                for k in range(self.__col):
                    total += self.__values[i][k] * m.__values[k][j]
                result.__values[i][j] = total
        return result

    cpdef Matrix elementProduct(self, Matrix m):
        """
        The elementProduct method takes a Matrix as an input and performs element wise multiplication. Puts result
        to the newly created Matrix. If the size of Matrix's row and column size does not match with the values
        list's row and column size, it throws MatrixDimensionMismatch exception.

        PARAMETERS
        ----------
        m : Matrix
            Matrix type input.

        RETURNS
        -------
        Matrix
            result Matrix.
        """
        cdef Matrix result
        cdef int i, j
        if self.__row != m.__row or self.__col != m.__col:
            raise MatrixDimensionMismatch
        result = Matrix(self.__row, self.__col)
        for i in range(self.__row):
            for j in range(self.__col):
                result.__values[i][j] = self.__values[i][j] * m.__values[i][j]
        return result

    cpdef double sumOfElements(self):
        """
        The sumOfElements method accumulates all the items in values list and
        returns this summation.

        RETURNS
        -------
        double
            sum of the items of values list.
        """
        cdef double total
        cdef int i
        total = 0.0
        for i in range(self.__row):
            total += sum(self.__values[i])
        return total

    cpdef double trace(self):
        """
        The trace method accumulates items of values list at the diagonal.

        RETURNS
        -------
        double
            sum of items at diagonal.
        """
        cdef double total
        cdef int i
        if self.__row != self.__col:
            raise MatrixNotSquare
        total = 0.0
        for i in range(self.__row):
            total += self.__values[i][i]
        return total

    cpdef Matrix transpose(self):
        """
        The transpose method creates a new Matrix, then takes the transpose of values list
        and puts transposition to the Matrix.

        RETURNS
        -------
        Matrix
            Matrix type output.
        """
        cdef Matrix result
        cdef int i, j
        result = Matrix(self.__col, self.__row)
        for i in range(self.__row):
            for j in range(self.__col):
                result.__values[j][i] = self.__values[i][j]
        return result

    cpdef Matrix partial(self,
                         int rowStart,
                         int rowEnd,
                         int colStart,
                         int colEnd):
        """
        The partial method takes 4 integer inputs; rowStart, rowEnd, colStart, colEnd and creates a Matrix size of
        rowEnd - rowStart + 1 x colEnd - colStart + 1. Then, puts corresponding items of values list
        to the new result Matrix.

        PARAMETERS
        ----------
        rowStart : int
            integer input for defining starting index of row.
        rowEnd : int
            integer input for defining ending index of row.
        colStart : int
            integer input for defining starting index of column.
        colEnd : int
            integer input for defining ending index of column.

        RETURNS
        -------
        Matrix
            result Matrix.
        """
        cdef Matrix result
        cdef int i, j
        result = Matrix(rowEnd - rowStart + 1, colEnd - colStart + 1)
        for i in range(rowStart, rowEnd + 1):
            for j in range(colStart, colEnd + 1):
                result.__values[i - rowStart][j - colStart] = self.__values[i][j]
        return result

    cpdef bint isSymmetric(self):
        """
        The isSymmetric method compares each item of values list at positions (i, j) with (j, i)
        and returns true if they are equal, false otherwise.

        RETURNS
        -------
        bool
            true if items are equal, false otherwise.
        """
        cdef int i, j
        if self.__row != self.__col:
            raise MatrixNotSquare
        for i in range(self.__row - 1):
            for j in range(self.__row):
                if self.__values[i][j] != self.__values[j][i]:
                    return False
        return True

    cpdef double determinant(self):
        """
        The determinant method first creates a new list, and copies the items of  values
        list into new list. Then, calculates the determinant of this
        new list.

        RETURNS
        -------
        double
            determinant of values list.
        """
        cdef double det
        cdef Matrix copyOfMatrix
        cdef int i, j, k
        if self.__row != self.__col:
            raise MatrixNotSquare
        det = 1.0
        copyOfMatrix = copy.deepcopy(self)
        for i in range(self.__row):
            det *= copyOfMatrix.__values[i][i]
            if det == 0.0:
                break
            for j in range(i + 1, self.__row):
                ratio = copyOfMatrix.__values[j][i] / copyOfMatrix.__values[i][i]
                for k in range(i, self.__col):
                    copyOfMatrix.__values[j][k] = copyOfMatrix.__values[j][k] - copyOfMatrix.__values[i][k] * ratio
        return det

    cpdef inverse(self):
        """
        The inverse method finds the inverse of values list.
        """
        cdef Matrix b
        cdef int i, j, k, l, ll, irow, icol
        cdef list indxc, indxr, ipiv
        cdef double big, dum, pivinv
        if self.__row != self.__col:
            raise MatrixNotSquare
        b = Matrix(self.__row, self.__row)
        indxc = []
        indxr = []
        ipiv = []
        for j in range(self.__row):
            ipiv.append(0)
        for i in range(1, self.__row + 1):
            big = 0.0
            irow = -1
            icol = -1
            for j in range(1, self.__row + 1):
                if ipiv[j - 1] != 1:
                    for k in range(1, self.__row + 1):
                        if ipiv[k - 1] == 0:
                            if abs(self.__values[j - 1][k - 1]) >= big:
                                big = abs(self.__values[j - 1][k - 1])
                                irow = j
                                icol = k
            if irow == -1 or icol == -1:
                raise DeterminantZero
            ipiv[icol - 1] = ipiv[icol - 1] + 1
            if irow != icol:
                for l in range(1, self.__row + 1):
                    dum = self.__values[irow - 1][l - 1]
                    self.__values[irow - 1][l - 1] = self.__values[icol - 1][l - 1]
                    self.__values[icol - 1][l - 1] = dum
                for l in range(1, self.__row + 1):
                    dum = b.__values[irow - 1][l - 1]
                    b.__values[irow - 1][l - 1] = b.__values[icol - 1][l - 1]
                    b.__values[icol - 1][l - 1] = dum
            indxr.append(irow)
            indxc.append(icol)
            if self.__values[icol - 1][icol - 1] == 0:
                raise DeterminantZero
            pivinv = 1.0 / self.__values[icol - 1][icol - 1]
            self.__values[icol - 1][icol - 1] = 1.0
            for l in range(1, self.__row + 1):
                self.__values[icol - 1][l - 1] = self.__values[icol - 1][l - 1] * pivinv
            for l in range(1, self.__row + 1):
                b.__values[icol - 1][l - 1] = b.__values[icol - 1][l - 1] * pivinv
            for ll in range(1, self.__row + 1):
                if ll != icol:
                    dum = self.__values[ll - 1][icol - 1]
                    self.__values[ll - 1][icol - 1] = 0.0
                    for l in range(1, self.__row + 1):
                        self.__values[ll - 1][l - 1] = self.__values[ll - 1][l - 1] - self.__values[icol - 1][
                            l - 1] * dum
                    for l in range(1, self.__row + 1):
                        b.__values[ll - 1][l - 1] = b.__values[ll - 1][l - 1] - b.__values[icol - 1][l - 1] * dum
        for l in range(self.__row, 0, -1):
            if indxr[l - 1] != indxc[l - 1]:
                for k in range(1, self.__row + 1):
                    dum = self.__values[k - 1][indxr[l - 1] - 1]
                    self.__values[k - 1][indxr[l - 1] - 1] = self.__values[k - 1][indxc[l - 1] - 1]
                    self.__values[k - 1][indxc[l - 1] - 1] = dum

    cpdef Matrix choleskyDecomposition(self):
        """
        The choleskyDecomposition method creates a new Matrix and puts the Cholesky Decomposition of values Array
        into this Matrix. Also, it throws MatrixNotSymmetric exception if it is not symmetric and
        MatrixNotPositiveDefinite exception if the summation is negative.

        RETURNS
        -------
        Matrix
            Matrix type output.
        """
        cdef Matrix b
        cdef int i, j
        cdef double total
        if not self.isSymmetric():
            raise MatrixNotSymmetric
        b = Matrix(self.__row, self.__col)
        for i in range(self.__row):
            for j in range(i, self.__row):
                total = self.__values[i][j]
                for k in range(i - 1, -1, -1):
                    total -= self.__values[i][k] * self.__values[j][k]
                if i == j:
                    if total <= 0.0:
                        raise MatrixNotPositiveDefinite
                    b.__values[i][i] = math.sqrt(total)
                else:
                    b.__values[j][i] = total / b.__values[i][i]
        return b

    cpdef __rotate(self,
                   double s,
                   double tau,
                   int i,
                   int j,
                   int k,
                   int l):
        """
        The rotate method rotates values list according to given inputs.

        PARAMETERS
        ----------
        s : double
            double input.
        tau : double
            double input.
        i : int
            integer input.
        j : int
            integer input.
        k : int
            integer input.
        l : int
            integer input.
        """
        cdef double g, h
        g = self.__values[i][j]
        h = self.__values[k][l]
        self.__values[i][j] = g - s * (h + g * tau)
        self.__values[k][l] = h + s * (g - h * tau)

    def characteristics(self)-> list:
        """
        The characteristics method finds and returns a sorted list of Eigenvecto}s. And it throws
        MatrixNotSymmetric exception if it is not symmetric.

        RETURNS
        -------
        list
            A sorted list of Eigenvectors.
        """
        cdef Matrix matrix1, v
        cdef list d, b, z, result
        cdef double EPS, sm, threshold, g, h, t, theta, c, s, tau
        cdef int ip, i, iq
        if not self.isSymmetric():
            raise MatrixNotSymmetric
        matrix1 = copy.deepcopy(self)
        v = Matrix(self.__row, self.__row, 1.0)
        d = []
        b = []
        z = []
        EPS = 0.000000000000000001
        for ip in range(self.__row):
            b.append(matrix1.__values[ip][ip])
            d.append(matrix1.__values[ip][ip])
            z.append(0.0)
        for i in range(1, 51):
            sm = 0.0
            for ip in range(self.__row - 1):
                for iq in range(ip + 1, self.__row):
                    sm += abs(matrix1.__values[ip][iq])
            if sm == 0.0:
                break
            if i < 4:
                threshold = 0.2 * sm / (self.__row ** 2)
            else:
                threshold = 0.0
            for ip in range(self.__row - 1):
                for iq in range(ip + 1, self.__row):
                    g = 100.0 * abs(matrix1.__values[ip][iq])
                    if i > 4 and g <= EPS * abs(d[ip]) and g <= EPS * abs(d[iq]):
                        matrix1.__values[ip][iq] = 0.0
                    else:
                        if abs(matrix1.__values[ip][iq]) > threshold:
                            h = d[iq] - d[ip]
                            if g <= EPS * abs(h):
                                t = matrix1.__values[ip][iq] / h
                            else:
                                theta = 0.5 * h / matrix1.__values[ip][iq]
                                t = 1.0 / (abs(theta) + math.sqrt(1.0 + theta ** 2))
                                if theta < 0.0:
                                    t = -t
                            c = 1.0 / math.sqrt(1 + t ** 2)
                            s = t * c
                            tau = s / (1.0 + c)
                            h = t * matrix1.__values[ip][iq]
                            z[ip] -= h
                            z[iq] += h
                            d[ip] -= h
                            d[iq] += h
                            matrix1.__values[ip][iq] = 0.0
                            for j in range(ip):
                                matrix1.__rotate(s, tau, j, ip, j, iq)
                            for j in range(ip + 1, iq):
                                matrix1.__rotate(s, tau, ip, j, j, iq)
                            for j in range(iq + 1, self.__row):
                                matrix1.__rotate(s, tau, ip, j, iq, j)
                            for j in range(self.__row):
                                v.__rotate(s, tau, j, ip, j, iq)
            for ip in range(self.__row):
                b[ip] = b[ip] + z[ip]
                d[ip] = b[ip]
                z[ip] = 0.0
        result = []
        for i in range(self.__row):
            if d[i] > 0:
                result.append(Eigenvector(d[i], v.getColumnVector(i)))
        result.sort(key=lambda eigenvector: eigenvector.getEigenvalue(), reverse=True)
        return result

    def __repr__(self):
        return f"{self.__values}"



from Hmm.Hmm cimport Hmm
from Math.Vector cimport Vector


cdef class Hmm1(Hmm):

    cdef Vector __pi

    cpdef calculatePi(self, list observations)
    cpdef calculateTransitionProbabilities(self, list observations)
    cpdef Vector __logOfColumn(self, int column)
    cpdef list viterbi(self, list s)



cdef class DiscreteDistribution(dict):

    cdef double __sum

    cpdef addItem(self, str item)
    cpdef removeItem(self, str item)
    cpdef addDistribution(self, DiscreteDistribution distribution)
    cpdef removeDistribution(self, DiscreteDistribution distribution)
    cpdef double getSum(self)
    cpdef int getIndex(self, str item)
    cpdef bint containsItem(self, str item)
    cpdef str getItem(self, int index)
    cpdef int getValue(self, int index)
    cpdef int getCount(self, str item)
    cpdef str getMaxItem(self)
    cpdef str getMaxItemIncludeTheseOnly(self, list includeTheseOnly)
    cpdef double getProbability(self, str item)
    cpdef dict getProbabilityDistribution(self)
    cpdef double getProbabilityLaplaceSmoothing(self, str item)
    cpdef double entropy(self)



cdef class HmmState(object):

    def __init__(self,
                 state: object,
                 emissionProbabilities: dict):
        """
        A constructor of HmmState class which takes a State and emission probabilities as inputs and
        initializes corresponding class variable with these inputs.

        PARAMETERS
        ----------
        state : object
            Data for this state.
        emissionProbabilities : dict
            Emission probabilities for this state
        """
        self.state = state
        self.emission_probabilities = emissionProbabilities

    cpdef object getState(self):
        """
        Accessor method for the state variable.

        RETURNS
        -------
        object
            state variable.
        """
        return self.state

    cpdef double getEmitProb(self, object symbol):
        """
        getEmitProb method returns the emission probability for a specific symbol.

        PARAMETERS
        ----------
        symbol : object
            Symbol for which the emission probability will be get.

        RETURNS
        -------
        float
            Emission probability for a specific symbol.
        """
        if symbol in self.emission_probabilities:
            return self.emission_probabilities[symbol]
        else:
            return 0.0

    def __repr__(self):
        return f"{self.state} {self.emission_probabilities}"



import collections


cdef class LRUCache(object):

    def __init__(self, cacheSize: int):
        """
        A constructor of LRUCache class which takes cacheSize as input. It creates new CacheLinkedList and
        HashMap.

        PARAMETERS
        ----------
        cacheSize : int
            Integer input defining cache size.
        """
        self.__cacheSize = cacheSize
        self.__map = collections.OrderedDict()

    cpdef bint contains(self, key: object):
        """
        The contains method takes a object type input key and returns true if the dictionary has the given key, false
        otherwise.

        PARAMETERS
        ----------
        key : object
            object type input key.

        RETURNS
        -------
            true if the HashMap has the given key, false otherwise.
        """
        return key in self.__map

    cpdef object get(self, key: object):
        """
        The get method takes object type input key and returns the least recently used value. First it checks whether the
        dictionary has the given key, if so it gets the corresponding cacheNode. It removes that cacheNode from
        LinkedList and adds it again to the beginning of the list since it is more likely to be used again. At the end,
        returns the data value of that cacheNode.

        PARAMETERS
        ----------
        key : object
            object type input key.

        RETURNS
        -------
        object
            data value if the dictionary has the given key, None otherwise.
        """
        if key in self.__map:
            value = self.__map[key]
            self.__map.pop(key)
            self.__map[key] = value
            return value
        else:
            return None

    cpdef add(self, key: object, data: object):
        """
        The add method take a key and a data as inputs. First it checks the size of the dictionary, if it is full (i.e
        equal to the given cacheSize) then it removes the last cacheNode in the list. If it has space for new entries,
        it creates new cacheNode with given inputs and adds this cacheNode to the linked list and also puts
        it to the dictionary.

        PARAMETERS
        ----------
        key : object
            object type input.
        data : object
            object type input
        """
        if len(self.__map) == self.__cacheSize:
            self.__map.popitem(last=False)
        self.__map[key] = data



cdef class Eigenvector(Vector):

    def __init__(self,
                 eigenvalue: double,
                 values: list):
        """
        A constructor of Eigenvector which takes a double eigenValue and an list values as inputs.
        It calls its super class Vector with values list and initializes eigenValue variable with its
        eigenValue input.

        PARAMETERS
        ----------
        eigenvalue : double
            eigenValue double input.
        values : list
            list input.
        """
        super().__init__(values)
        self.eigenvalue = eigenvalue

    cpdef float getEigenvalue(self):
        """
        The eigenValue method which returns the eigenValue variable.

        RETURNS
        -------
        double
            eigenValue variable.
        """
        return self.eigenvalue



from setuptools import setup

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")
from Cython.Build import cythonize

setup(
    ext_modules=cythonize(["Hmm/*.pyx"],
                          compiler_directives={'language_level': "3"}),
    name='NlpToolkit-Hmm-Cy',
    version='1.0.4',
    packages=['Hmm'],
    package_data={'Hmm': ['*.pxd', '*.pyx', '*.c']},
    url='https://github.com/StarlangSoftware/Hmm-Cy',
    license='',
    author='olcaytaner',
    author_email='olcay.yildiz@ozyegin.edu.tr',
    description='Hidden Markov Model Library',
    install_requires=['NlpToolkit-Math-Cy', 'NlpToolkit-DataStructure-Cy'],
    long_description=long_description,
    long_description_content_type='text/markdown'
)



cdef class CounterHashMap(dict):

    def __init__(self, **kwargs):
        """
        A constructor which calls its super.
        """
        super().__init__(kwargs)

    def __str__(self)-> str:
        """
        The __str__ method loops through the mappings contained in this map and returns the string of each entry's key
        and value.

        RETURNS
        -------
        str
            String of the each entry's key and value.
        """
        result = ""
        for key in self.keys():
            result += key + " " + self[key].__str__() + "\n"
        return result

    cpdef put(self, object key):
        """
        The put method takes a object type input. If this map contains a mapping for the key, it puts this key after
        incrementing its value by one. If his map does not contain a mapping, then it directly puts key with the value
        of 1.

        PARAMETERS
        ----------
        key : object
            key to put.
        """
        if key in self:
            self[key] = self[key] + 1
        else:
            self[key] = 1

    cpdef putNTimes(self, object key, int N):
        """
        The putNTimes method takes an object and an integer N as inputs. If this map contains a mapping for the key, it
        puts this key after incrementing its value by N. If his map does not contain a mapping, then it directly puts
        key with the value of N.

        PARAMETERS
        ----------
        key : object
            key to put.
        N : int
            to increment value.
        """
        if key in self:
            self[key] += N
        else:
            self[key] = N

    cpdef int count(self, object key):
        """
        The count method takes an object as input, if this map contains a mapping for the key, it returns the value
        corresponding this key, 0 otherwise.

        PARAMETERS
        ----------
        key : object
            key to get value.

        RETURNS
        -------
        int
            the value corresponding given key, 0 if it is not mapped.
        """
        if key in self:
            return self[key]
        else:
            return 0

    cpdef int sumOfCounts(self):
        """
        The sumOfCounts method loops through the values contained in this map and accumulates the counts of this values.

        RETURNS
        -------
        int
            accumulated counts.
        """
        cdef int total
        total = 0
        for key in self:
            total += self[key]
        return total

    cpdef object max(self, double threshold = 0.0):
        """
        The max method takes a threshold as input and loops through the mappings contained in this map.
        It accumulates the count values and if the current entry's count value is greater than maxCount, which is
        initialized as 0, it updates the maxCount as current count and maxKey as the current count's key.
        At the end of the loop, if the ratio of maxCount/total is greater than the given threshold it returns maxKey,
        else None.

        PARAMETERS
        ----------
        threshold : float
            threshold float value.

        RETURNS
        -------
        object
            object type maxKey if greater than the given threshold, None otherwise.
        """
        cdef int maxCount, total
        maxCount = 0
        total = 0
        maxKey = None
        for key in self:
            total += self[key]
            if self[key] > maxCount:
                maxCount = self[key]
                maxKey = key
        if maxCount / total > threshold:
            return maxKey
        else:
            return None

    cpdef add(self, CounterHashMap toBeAdded):
        """
        The add method adds value of each key of toBeAdded to the current counterHashMap.

        PARAMETERS
        ----------
        toBeAdded : CounterHashMap
            CounterHashMap to be added to this counterHashMap.
        """
        for value in toBeAdded:
            self.putNTimes(value, toBeAdded[value])

    cpdef list topN(self, int N):
        """
        The topN method takes an integer N as inout. It creates an list result and loops through the the
        mappings contained in this map and adds each entry to the result list. Then sort this list
        according to their values and returns a list which is a sublist of result with N elements.

        PARAMETERS
        ----------
        N : int
            integer value for defining size of the sublist.

        RETURNS
        -------
        list
            a sublist of N element.
        """
        cdef list result
        result = []
        for key in self:
            result.append([self[key], key])
        result.sort(reverse=True)
        return result[0:N]



cdef class LRUCache(object):

    cdef int __cacheSize
    cdef object __map

    cpdef bint contains(self, key: object)
    cpdef object get(self, key: object)
    cpdef add(self, key: object, data: object)